from typing import   Self
from pydantic import ConfigDict, model_validator
from labfreed.labfreed_infrastructure import LabFREED_BaseModel, LabFREED_ValidationError, ValidationMsgLevel
from labfreed.pac_id.pac_id import PAC_ID


class AttributeRequestPayload(LabFREED_BaseModel):
    model_config = ConfigDict(frozen=True)
    
    pac_ids: list[str]
    language_preferences: list[str]|None = None
    restrict_to_attribute_groups: list[str]|None = None
    suppress_forward_lookup: bool = False
    
    def as_json(self):
        return self.model_dump_json()
    
    @classmethod
    def from_json(cls, json):
        return cls.model_validate_json(json)
    
    @model_validator(mode="before")
    @classmethod
    def _handle_single_pac_url(cls, data):
        p = data.get('pac_urls')
        if isinstance(p, str):
            data['pac_urls'] = [p]
        return data
    
    @model_validator(mode="after")
    def _validate_pacs(self) -> Self:
        if len(self.pac_ids) > 100:
            self._add_validation_message(
                        source="pacs",
                        level = ValidationMsgLevel.ERROR,
                        msg='The number of pac-ids must be limited to 100'
                    )
            
        for pac_url in self.pac_ids:
            try:
                PAC_ID.from_url(pac_url)
            except LabFREED_ValidationError:
                self._add_validation_message(
                        source="pacs",
                        level = ValidationMsgLevel.ERROR,
                        msg='{pac_url} is not a valid PAC-ID'
                    )
                
        if not self.is_valid:
            raise LabFREED_ValidationError(message='Invalid request', validation_msgs=self.validation_messages())
                
        return self
    
    
    





from abc import ABC
from datetime import  datetime
import re
from typing import Annotated, Any,  Literal, Union, get_args
from urllib.parse import urlparse

from labfreed.utilities.ensure_utc_time import ensure_utc
from labfreed.labfreed_infrastructure import  LabFREED_BaseModel, ValidationMsgLevel, _quote_texts
from pydantic import   Field, field_validator, model_validator


class AttributeBase(LabFREED_BaseModel, ABC):
    key: str
    value: Any
    label: str = ""
        
    def __init__(self, **data):
        # Automatically inject the Literal value for `type`
        discriminator_value = self._get_discriminator_value()
        data["type"] = discriminator_value
        super().__init__(**data)       
    
    @classmethod
    def _get_discriminator_value(cls) -> str:
        """Extract the Literal value from the 'type' annotation."""
        try:
            type_annotation = cls.__annotations__["type"]
            literal_value = get_args(type_annotation)[0]
            return literal_value
        except Exception as e:
            raise TypeError(
                f"{cls.__name__} must define `type: Literal[<value>]` annotation"
            ) from e
        
    
class DateTimeAttribute(AttributeBase):
    type: Literal["datetime"] 
    value: datetime
    
    @field_validator('value', mode='after')
    def set_utc__if_naive(cls, value):
        if isinstance(value, datetime):
            return ensure_utc(value)
        else:
            return value
        
class DateTimeListAttribute(AttributeBase):
    type: Literal["datetime-list"] 
    value: list[datetime]
    
    @field_validator('value', mode='after')
    def set_utc__if_naive(cls, value):
        value_out = []
        for v in value:
            if isinstance(v, datetime):
                value_out.append(ensure_utc(v))
            else:
                raise ValueError(f'{v} is of type {type(v)}. It must be datetime')
        return value_out
    
    
    
class BoolAttribute(AttributeBase):
    type: Literal["bool"] 
    value: bool
    
class BoolListAttribute(AttributeBase):
    type: Literal["bool-list"] 
    value: list[bool]



    
class TextAttribute(AttributeBase):
    type: Literal["text"] 
    value: str
    
    @model_validator(mode='after')
    def _validate_value(self):
        _validate_text(self, self.value)
        return self
    
class TextListAttribute(AttributeBase):
    type: Literal["text-list"] 
    value: list[str]
    
    @model_validator(mode='after')
    def _validate_value(self):
        l = [self.value] if isinstance(self.value, str) else self.value
        for v in l:
            _validate_text(self, v)
        return self
    

def _validate_text(mdl:LabFREED_BaseModel, v):
    if len(v) > 5000: 
        mdl._add_validation_message(
            source="Text Attribute",
            level=ValidationMsgLevel.WARNING,  # noqa: F821
            msg=f"Text attribute {v} exceeds 5000 characters. It is recommended to stay below",
            highlight_pattern = f'{v}'
        )
            


class ReferenceAttribute(AttributeBase):
    type: Literal["reference"]
    value: str 
    
class ReferenceListAttribute(AttributeBase):
    type: Literal["reference-list"]
    value: list[str]
    
    
        

class ResourceAttribute(AttributeBase):
    type: Literal["resource"]
    value: str 
    
    @model_validator(mode='after')
    def _validate_value(self):
        _validate_resource(self, self.value)
        return self

class ResourceListAttribute(AttributeBase):
    type: Literal["resource-list"]
    value: list[str]
    
    @model_validator(mode='after')
    def _validate_value(self):
        value_list = self.value if isinstance(self.value, list) else [self.value]
        for v in value_list:
            _validate_resource(self, v)
        return self
    
def _validate_resource(mdl:LabFREED_BaseModel, v):
    r = urlparse(v)
    if not all([r.scheme, r.netloc]):
        mdl._add_validation_message(
            source="Resource Attribute",
            level=ValidationMsgLevel.ERROR,  # noqa: F821
            msg="Must be a valid url",
            highlight_pattern = f'{v}'
        )
    pattern = re.compile(r"\.\w{1,3}$", re.IGNORECASE)
    if not bool(pattern.search(v)):
        mdl._add_validation_message(
            source="Resource Attribute",
            level=ValidationMsgLevel.WARNING,  # noqa: F821
            msg="It is RECOMMENDED resource links end with a file extension",
            highlight_pattern = f'{v}'
        )
    
        
        
class NumericValue(LabFREED_BaseModel):
    numerical_value: str
    unit: str
    
    @model_validator(mode='after')
    def _validate_value(self):
        value = self.numerical_value
        if not_allowed_chars := set(re.sub(r'[0-9\.\-\+Ee]', '', value)):
            self._add_validation_message(
                source="Numeric Attribute",
                level=ValidationMsgLevel.ERROR,  # noqa: F821
                msg=f"Characters {_quote_texts(not_allowed_chars)} are not allowed in quantity segment. Must be a number.",
                highlight_pattern = f'{value}',
                highlight_sub=not_allowed_chars
            )
        if not re.fullmatch(r'-?\d+(\.\d+)?([Ee][\+-]?\d+)?', value):
            self._add_validation_message(
                source="Numeric Attribute",
                level=ValidationMsgLevel.ERROR,
                msg=f"{value} cannot be converted to number",
                highlight_pattern = f'{value}'               
            )
        return self
    
    @model_validator(mode="after")
    def _validate_units(self):
        '''A sanity check on unit complying with UCUM. NOTE: It is not a complete validation
        - I check for blankspaces and ^, which are often used for units, but are invalid.
        - the general structure of a ucum unit is validated, but 1)parentheses are not matched 2) units are not validated 3)prefixes are not checked
        '''
        if ' ' in self.unit or '^' in self.unit:
            self._add_validation_message(
                    source="Numeric Attribute",
                    level= ValidationMsgLevel.ERROR,
                    msg=f"Unit {self.unit} is invalid. Must not contain blankspace  or '^'.",
                    highlight_pattern = self.unit
            )
        elif not re.fullmatch(r"^(((?P<unit>[\w\[\]]+?)(?P<exponent>\-?\d+)?|(?P<annotation>)\{\w+?\})(?P<operator>[\./]?)?)+", self.unit):
            self._add_validation_message(
                    source="Numeric Attribute",
                    level= ValidationMsgLevel.WARNING,
                    msg=f"Unit {self.unit} is probably invalid. Ensure it complies with UCUM specifications.",
                    highlight_pattern = self.unit
            )
        return self
    
    


class NumericAttribute(AttributeBase):
    type: Literal["numeric"] 
    value: NumericValue 
    
class NumericListAttribute(AttributeBase):
    type: Literal["numeric-list"] 
    value: list[NumericValue]
    
    
    
class ObjectAttribute(AttributeBase):
    type: Literal["object"] 
    value: dict[str, Any]
    
    
class ObjectListAttribute(AttributeBase):
    type: Literal["object-list"] 
    value: list[dict[str, Any]]
           

     
     
Attribute = Annotated[
    Union[
        ReferenceAttribute,
        DateTimeAttribute,
        BoolAttribute,
        TextAttribute,
        NumericAttribute,
        ResourceAttribute,
        ObjectAttribute,
        
        ReferenceListAttribute,
        DateTimeListAttribute,
        BoolListAttribute,
        TextListAttribute,
        NumericListAttribute,
        ResourceListAttribute,
        ObjectListAttribute
    ],
    Field(discriminator="type")
]


class AttributeGroup(LabFREED_BaseModel):
    key: str
    label: str = ""
    attributes: list[Attribute]
       


class AttributesOfPACID(LabFREED_BaseModel):
    pac_id: str
    attribute_groups: list[AttributeGroup]
    
    

class AttributeResponsePayload(LabFREED_BaseModel):
    schema_version: str = Field(default='1.0')
    language:str 
    pac_attributes: list[AttributesOfPACID]      
    
    def to_json(self):
        return self.model_dump_json(exclude_none=True)
    
    












from datetime import UTC, date, datetime, time
import json
from typing import  Literal
from enum import Enum
import warnings
from pydantic import RootModel, field_validator

from labfreed.labfreed_infrastructure import LabFREED_BaseModel
from labfreed.pac_attributes.api_data_models.response import AttributeBase, AttributeGroup, BoolAttribute, BoolListAttribute, DateTimeAttribute, DateTimeListAttribute,  NumericAttribute, NumericListAttribute, NumericValue, ObjectAttribute,ObjectListAttribute, ReferenceAttribute, ReferenceListAttribute, ResourceAttribute, ResourceListAttribute, TextAttribute, TextListAttribute
from labfreed.pac_attributes.client.attribute_cache import CacheableAttributeGroup
from labfreed.pac_id.pac_id import PAC_ID
from labfreed.trex.pythonic.quantity import Quantity



class pyReference(RootModel[str]):

    def __str__(self):
        return str(self.root)
    
class pyResource(RootModel[str]):

    def __str__(self):
        return str(self.root)


# the allowed scalar types
AllowedValue = str | bool | datetime | pyReference | pyResource | Quantity | int | float | dict | object
# homogeneous list of those
AllowedList = list[AllowedValue]

class pyAttribute(LabFREED_BaseModel):
    key:str
    label:str = ""
    value: AllowedValue | AllowedList
    
    @property
    def value_list(self):    
        '''helper function to more conveniently iterate over value elements, even if it's scalar'''   
        return self.value if isinstance(self.value, list) else [self.value]
    
    
    @field_validator('value', mode='before')
    def handle_one_element_list(v):
        if isinstance(v, list) and len(v)==1:
            return v[0]
        else:
            return v
        
    @field_validator('key', mode='before')
    def handle_enum_key(v):
        if isinstance(v, Enum) :
            return v.value
        else:
            return v
        
    

class pyAttributes(RootModel[list[pyAttribute]]):
    def to_payload_attributes(self) -> list[AttributeBase]:
        out = []
        for e in self.root:
            apt = self._attribute_to_attribute_payload_type(e)
            if isinstance(apt.value, list) and len(apt.value) ==1:
                apt.value = apt.value[0]
            out.append(apt)
        return out
    
            
    @staticmethod        
    def _attribute_to_attribute_payload_type(attribute:pyAttribute) -> AttributeBase:
        common_args = {
            "key": attribute.key,
            "label": attribute.label,
        }
        value_list = attribute.value_list
        first_value = value_list[0]
        if isinstance(first_value, bool):
            if len(value_list) == 1:
                return BoolAttribute(value=value_list[0], **common_args)
            else:
                return BoolListAttribute(value=value_list, **common_args)
         
            
        elif isinstance(first_value, datetime | date | time):
            for v in value_list:
                if not v.tzinfo:
                    warnings.warn(f'No timezone given for {v}. Assuming it is in UTC.')
                    v.replace(tzinfo=UTC)
            if len(value_list) == 1:
                return DateTimeAttribute(value=value_list[0], **common_args)
            else:
                return DateTimeListAttribute(value=value_list, **common_args)
            # return DateTimeAttribute(value =_date_value_from_python_type(value).value, **common_args)
            
        
        elif isinstance(first_value, Quantity|int|float):
            values = []
            for v in value_list:
                if not isinstance(v, Quantity):
                    v = Quantity(value=v, unit='dimensionless')
                values.append(NumericValue(numerical_value=v.value_as_str(), 
                                            unit = v.unit))
            if len(values) == 1:
                num_attr = NumericAttribute(value=values[0], **common_args)
            else:
                num_attr = NumericListAttribute(value=values, **common_args)
            num_attr.print_validation_messages()
            
            return num_attr
        
        elif isinstance(first_value, str):
            # capture quantities in the form of "100.0e5 g/L"
            if Quantity.can_convert_to_quantity(first_value):
                values = []
                for v in value_list:
                    q = Quantity.from_str_with_unit(v)
                    values.append( NumericValue(numerical_value=q.value_as_str(), unit = q.unit) )
                if len(values) == 1:
                    return NumericAttribute(value=values[0], **common_args)
                else:
                    return NumericListAttribute(value=values, **common_args)

            else:
                if len(value_list) == 1:
                    return TextAttribute(value=value_list[0], **common_args)
                else:
                    return TextListAttribute(value=value_list, **common_args)
            
        elif isinstance(first_value, pyReference):
            values = [v.root for v in value_list]
            if len(values) == 1:
                return ReferenceAttribute(value=values[0], **common_args)
            else:
                return ReferenceListAttribute(value=values, **common_args)
        
        elif isinstance(first_value, pyResource):
            values = [v.root for v in value_list]
            if len(values) == 1:
                return ResourceAttribute(value=values[0], **common_args)
            else:
                return ResourceListAttribute(value=values, **common_args)
            
        elif isinstance(first_value, PAC_ID):
            values = [v.to_url(include_extensions=False) for v in value_list]
            if len(values) == 1:
                return ReferenceAttribute(value=values[0], **common_args)
            else:
                return ReferenceListAttribute(value=values, **common_args)
        
        else: #this covers the last resort case of arbitrary objects. Must be json serializable.
            try :
                values = [json.loads(json.dumps(v)) for v in value_list]
                if len(values) == 1:
                    return ObjectAttribute(value=values[0], **common_args)
                else:
                    return ObjectListAttribute(value=values, **common_args)
            except TypeError as e:  # noqa: F841
                raise ValueError(f'Invalid Type: {type(first_value)} cannot be converted to attribute. You may want to use ObjectAttribute, but would have to implement the conversion from your python type yourself.')
        
        
        
    @staticmethod
    def from_payload_attributes(attributes:list[AttributeBase]) -> 'pyAttributes':
        out = list()
        for a in attributes:
            value_list = a.value if isinstance(a.value, list) else [a.value]
            match a:
                case ReferenceAttribute() | ReferenceListAttribute():
                    values =  [pyReference(v) for v in value_list]
                    
                case ResourceAttribute() | ResourceListAttribute():
                    values =  [pyResource(v) for v in value_list]
                    
                case NumericAttribute() | NumericListAttribute():                                       
                    values = [ Quantity.from_str_value(value=v.numerical_value, unit=v.unit) for v in value_list]

                case BoolAttribute() | BoolListAttribute():
                    values = value_list
                    
                case TextAttribute() | TextListAttribute():
                    values = value_list
                    
                case DateTimeAttribute() | DateTimeListAttribute():                    
                    values = value_list
                
                case ObjectAttribute() | ObjectListAttribute():
                    values = value_list

                       
            attr = pyAttribute(key=a.key, 
                               label=a.label,
                               value=values
            )
            out.append(attr )
        return out
            
            
        
class pyAttributeGroup(CacheableAttributeGroup):
    attributes:dict[str,pyAttribute]
    
    @staticmethod
    def from_attribute_group(attribute_group:AttributeGroup):
        data = vars(attribute_group).copy()
        data["attributes"] = {a.key: a for a in pyAttributes.from_payload_attributes(attribute_group.attributes)}
        return pyAttributeGroup(**data)




from datetime import UTC, datetime, timedelta
from typing import Literal, Protocol


from labfreed.pac_attributes.api_data_models.response import AttributeGroup
from labfreed.pac_id.pac_id import PAC_ID



class CacheableAttributeGroup(AttributeGroup):
    origin:str
    language:str
    value_from: datetime | None = None  
    
    def still_valid(self, accept_cache_for_minutes):
        if self.value_from is None:
            return False
        else:
            return ( datetime.now(tz=UTC) - timedelta(minutes=accept_cache_for_minutes)) > self.value_from 






class AttributeCache(Protocol):
    def get_all(self, service_url:str, pac:PAC_ID) -> list[CacheableAttributeGroup]:
        pass
    
    def get_attribute_groups(self, service_url:str, pac:PAC_ID, attribute_groups:list[str]):
        pass
        
    def update(self, service_url:str, pac:PAC_ID, attribute_groups:list[CacheableAttributeGroup]):
        pass
    

            
class MemoryAttributeCache(AttributeCache):
    '''simple in-memory implementation of AttributeCache'''
    def __init__(self):
        self._store = dict()
    
    def get_all(self, service_url:str, pac:PAC_ID|str) -> list[CacheableAttributeGroup]:
        if isinstance(pac, str):
            pac = PAC_ID.from_url(pac)
        k = self._generate_dict_key(service_url=service_url, pac=pac)
        
        ags = [CacheableAttributeGroup.model_validate(e) for e in self._store.get(k, [])]
        return ags
    
    def get_attribute_groups(self, service_url:str, pac:PAC_ID, attribute_groups:list[str]):
        all_ags = self.get_all(service_url=service_url, pac=pac)
        selected_ags = [ag for ag in all_ags if ag.key in attribute_groups]
        return selected_ags
        
    def update(self, service_url:str, pac:PAC_ID, attribute_groups: list[CacheableAttributeGroup] ):
        k = self._generate_dict_key(service_url=service_url, pac=pac)
        self._store.update({k: [e.model_dump() for e in attribute_groups]})
    
    @staticmethod
    def _generate_dict_key(service_url:str, pac:PAC_ID):
        key = service_url +";"+ pac.to_url(include_extensions=False) 
        return key